/*
 * appstream-data-appdata-converter is an application for modifying appdata files
 * Copyright (C) 2019 Aleksei Nikiforov <darktemplar@basealt.ru>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include <stdio.h>

#include <string>
#include <fstream>
#include <sstream>
#include <list>
#include <map>
#include <regex>

#include <boost/program_options.hpp>

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/xml_parser.hpp>
#include <boost/property_tree/ini_parser.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>

void trim_children(boost::property_tree::ptree &propertyTree)
{
	for (auto iter = propertyTree.begin(); iter != propertyTree.end(); ++iter)
	{
		// only trim elements if they become empty after trimming
		auto value = iter->second.get_value<std::string>();
		boost::algorithm::trim(value);

		if (value.empty())
		{
			iter->second.put_value(value);
		}

		trim_children(iter->second);
	}
}

std::map<std::string, boost::property_tree::ptree> convert_children(const boost::property_tree::ptree &propertyTree, std::string lang = std::string())
{
	std::map<std::string, boost::property_tree::ptree> result;

	auto value = propertyTree.get_value<std::string>();
	auto trimmed_value = boost::algorithm::trim_copy(value);

	auto lang_child = propertyTree.get_child_optional("<xmlattr>.xml:lang");
	if (lang_child)
	{
		auto lang_value = lang_child->get_value<std::string>();
		boost::algorithm::trim(lang_value);

		if (!lang_value.empty())
		{
			lang = lang_value;
		}
	}

	if (!trimmed_value.empty())
	{
		result[lang] = boost::property_tree::ptree(value);
	}

	for (auto iter = propertyTree.begin(); iter != propertyTree.end(); ++iter)
	{
		if (iter->first == "<xmlattr>")
		{
			continue;
		}

		auto children_map = convert_children(iter->second, lang);
		for (auto children_iter = children_map.begin(); children_iter != children_map.end(); ++children_iter)
		{
			result[children_iter->first].add_child(iter->first, children_iter->second);
		}
	}

	return result;
}

int main(int argc, char **argv)
{
	std::string input_xml_file;
	std::string output_xml_file;
	std::string input_desktop_file;
	std::string pkgname;
	std::vector<std::string> releases_vector;
	std::vector<std::string> icons_vector;
	std::vector<std::string> exclude_lang_vector;

	boost::program_options::options_description desc("General Options");
	desc.add_options()
		("help,h", "produce help message")
		("release,r", boost::program_options::value< std::vector<std::string> >(&releases_vector), "add release information to output. format is 'timestamp;version'")
		("icon,i", boost::program_options::value< std::vector<std::string> >(&icons_vector), "add icon information to output. format is 'type;width;height;filename'")
		("exclude_language,e", boost::program_options::value< std::vector<std::string> >(&exclude_lang_vector), "don't include information in specific language")
	;

	boost::program_options::options_description hidden("Positional Options");
	hidden.add_options()
		("input_xml_file", boost::program_options::value<std::string>(&input_xml_file)->required(), "Input appdata.xml file")
		("output_xml_file", boost::program_options::value<std::string>(&output_xml_file)->required(), "Output appdata.xml file")
		("input_desktop_file", boost::program_options::value<std::string>(&input_desktop_file)->required(), "Input desktop file")
		("pkgname", boost::program_options::value<std::string>(&pkgname)->required(), "Name of package containing this appdata.xml file")
	;

	boost::program_options::options_description all_options;
	all_options.add(desc);
	all_options.add(hidden);

	boost::program_options::positional_options_description pos_opts;

	pos_opts.add("input_xml_file", 1);
	pos_opts.add("output_xml_file", 1);
	pos_opts.add("input_desktop_file", 1);
	pos_opts.add("pkgname", 1);

	boost::program_options::variables_map vm;

	try
	{
		boost::program_options::store(boost::program_options::command_line_parser(argc, argv).options(all_options).positional(pos_opts).run(), vm);
	}
	catch (const std::exception &e)
	{
		fprintf(stderr, "Error: %s\n", e.what());
		return -1;
	}

	if (vm.count("help"))
	{
		std::stringstream desc_stream;
		desc_stream << all_options;

		printf("%s\n", desc_stream.str().c_str());
		return 1;
	}

	try
	{
		boost::program_options::notify(vm);
	}
	catch (const std::exception &e)
	{
		fprintf(stderr, "Error: %s\n", e.what());
		return -1;
	}

	boost::property_tree::ptree propertyTree;

	try
	{
		std::ifstream stream(input_xml_file);
		boost::property_tree::read_xml(stream, propertyTree);
	}
	catch (const std::exception &e)
	{
		fprintf(stderr, "Failed to open file to read: %s\nError: %s\n", input_xml_file.c_str(), e.what());
		return -1;
	}

	trim_children(propertyTree);

	// Convert application to component
	{
		auto application_element = propertyTree.get_child_optional("application");
		if (application_element)
		{
			auto id_children = application_element->equal_range("id");
			for (auto iter = id_children.first; iter != id_children.second; ++iter)
			{
				std::string type;

				{
					auto type_child = iter->second.get_child_optional("<xmlattr>.type");
					if (type_child)
					{
						type = type_child->get_value<std::string>();
					}
				}

				iter->second.erase("<xmlattr>");
				
				if (!type.empty())
				{
					application_element->put_child("<xmlattr>.type", boost::property_tree::ptree(type));
				}
			}

			propertyTree.put_child("component", *application_element);
		}
	}

	propertyTree.erase("application");

	auto component_tree = propertyTree.get_child_optional("component");
	if (!component_tree)
	{
		fprintf(stderr, "Error: file doesn't have proper root element: %s\n", argv[1]);
		return -1;
	}

	// Find all description parts and save them in corresponding parts of description map
	// NOTE: currently can only split description if only one 'description' tag is present,
	// otherwise it's left as it's present
	if (component_tree->count("description") == 1)
	{
		auto description_child = component_tree->get_child("description");
		std::map<std::string, boost::property_tree::ptree> results = convert_children(description_child);

		// Finished processing description fields, now remove it
		component_tree->erase("description");

		{
			auto iter = results.find(std::string());
			if (iter != results.end())
			{
				component_tree->add_child("description", iter->second);

				results.erase(iter);
			}
		}

		for (auto iter = results.begin(); iter != results.end(); ++iter)
		{
			if (std::find(exclude_lang_vector.begin(), exclude_lang_vector.end(), iter->first) != exclude_lang_vector.end())
			{
				continue;
			}

			boost::property_tree::ptree descr_tree = iter->second;
			descr_tree.put_child("<xmlattr>.xml:lang", boost::property_tree::ptree(iter->first));
			component_tree->add_child("description", descr_tree);
		}
	}

	// Process pkgname
	{
		component_tree->erase("pkgname");

		component_tree->put_child("pkgname", boost::property_tree::ptree(pkgname));
	}

	// Process releases
	{
		component_tree->erase("releases");

		if (not releases_vector.empty())
		{
			boost::property_tree::ptree releases_tree;

			for (auto iter = releases_vector.begin(); iter != releases_vector.end(); ++iter)
			{
				boost::property_tree::ptree release_item_tree;

				std::vector<std::string> split_vector;
				boost::algorithm::split(split_vector, *iter, boost::is_any_of(";"));

				if (split_vector.size() != 2)
				{
					fprintf(stderr, "Invalid argument: release = %s\n", iter->c_str());
					continue;
				}

				release_item_tree.put_child("<xmlattr>.timestamp", boost::property_tree::ptree(split_vector[0]));
				release_item_tree.put_child("<xmlattr>.version", boost::property_tree::ptree(split_vector[1]));

				releases_tree.add_child("release", release_item_tree);
			}

			component_tree->put_child("releases", releases_tree);
		}
	}

	// Process icons
	{
		component_tree->erase("icon");

		for (auto iter = icons_vector.begin(); iter != icons_vector.end(); ++iter)
		{
			boost::property_tree::ptree icon_item_tree;

			std::vector<std::string> split_vector;
			boost::algorithm::split(split_vector, *iter, boost::is_any_of(";"));

			if (split_vector.size() != 4)
			{
				fprintf(stderr, "Invalid argument: icon = %s\n", iter->c_str());
				continue;
			}

			icon_item_tree.put_child("<xmlattr>.type", boost::property_tree::ptree(split_vector[0]));
			icon_item_tree.put_child("<xmlattr>.width", boost::property_tree::ptree(split_vector[1]));
			icon_item_tree.put_child("<xmlattr>.height", boost::property_tree::ptree(split_vector[2]));

			icon_item_tree.put_value(split_vector[3]);

			component_tree->add_child("icon", icon_item_tree);
		}
	}

	// Process data from desktop, if it's missing in xml file
	{
		boost::property_tree::ptree desktopPropertyTree;

		try
		{
			std::ifstream stream(input_desktop_file);
			boost::property_tree::read_ini(stream, desktopPropertyTree);
		}
		catch (const std::exception &e)
		{
			fprintf(stderr, "Failed to open file to read: %s\nError: %s\n", input_desktop_file.c_str(), e.what());
			return -1;
		}

		auto desktop_entry = desktopPropertyTree.get_child_optional("Desktop Entry");
		if (desktop_entry)
		{
			if (component_tree->find("name") == component_tree->not_found())
			{
				auto ini_element = desktop_entry->get_child_optional("Name");
				if (ini_element)
				{
					component_tree->add_child("name", *ini_element);

					// Also check translated items
					std::regex translated_regex("^Name\\[([^\\]]+)\\]$");
					std::smatch results;

					for (auto iter = desktop_entry->begin(); iter != desktop_entry->end(); ++iter)
					{
						if (std::regex_match(iter->first, results, translated_regex))
						{
							if (std::find(exclude_lang_vector.begin(), exclude_lang_vector.end(), results[1].str()) != exclude_lang_vector.end())
							{
								continue;
							}

							boost::property_tree::ptree child(iter->second.get_value<std::string>());
							child.put_child("<xmlattr>.xml:lang", boost::property_tree::ptree(results[1].str()));
							component_tree->add_child("name", child);
						}
					}
				}
			}

			if (component_tree->find("summary") == component_tree->not_found())
			{
				auto ini_element = desktop_entry->get_child_optional("Comment");
				if (ini_element)
				{
					component_tree->put_child("summary", *ini_element);

					// Also check translated items
					std::regex translated_regex("^Comment\\[([^\\]]+)\\]$");
					std::smatch results;

					for (auto iter = desktop_entry->begin(); iter != desktop_entry->end(); ++iter)
					{
						if (std::regex_match(iter->first, results, translated_regex))
						{
							if (std::find(exclude_lang_vector.begin(), exclude_lang_vector.end(), results[1].str()) != exclude_lang_vector.end())
							{
								continue;
							}

							boost::property_tree::ptree child(iter->second.get_value<std::string>());
							child.put_child("<xmlattr>.xml:lang", boost::property_tree::ptree(results[1].str()));
							component_tree->add_child("summary", child);
						}
					}
				}
			}

			if (component_tree->find("keywords") == component_tree->not_found())
			{
				auto ini_element = desktop_entry->get_child_optional("Keywords");
				if (ini_element)
				{
					std::vector<std::string> split_vector;
					auto tmpvalue = ini_element->get_value<std::string>();
					boost::algorithm::split(split_vector, tmpvalue, boost::is_any_of(";"));

					bool found = false;
					boost::property_tree::ptree subitem_tree;

					for (auto iter = split_vector.begin(); iter != split_vector.end(); ++iter)
					{
						if (!iter->empty())
						{
							found = true;
							subitem_tree.add_child("keyword", boost::property_tree::ptree(*iter));
						}
					}

					if (found)
					{
						component_tree->put_child("keywords", subitem_tree);
					}
				}
			}

			if (component_tree->find("categories") == component_tree->not_found())
			{
				auto ini_element = desktop_entry->get_child_optional("Categories");
				if (ini_element)
				{
					std::vector<std::string> split_vector;
					auto tmpvalue = ini_element->get_value<std::string>();
					boost::algorithm::split(split_vector, tmpvalue, boost::is_any_of(";"));

					bool found = false;
					boost::property_tree::ptree subitem_tree;

					for (auto iter = split_vector.begin(); iter != split_vector.end(); ++iter)
					{
						if (!iter->empty())
						{
							found = true;
							subitem_tree.add_child("category", boost::property_tree::ptree(*iter));
						}
					}

					if (found)
					{
						component_tree->put_child("categories", subitem_tree);
					}
				}
			}
		}
	}

	try
	{
		std::ofstream out_stream(output_xml_file);
		boost::property_tree::write_xml(out_stream, propertyTree, boost::property_tree::xml_writer_make_settings<std::string>('\t', 1));
	}
	catch (const std::exception &e)
	{
		fprintf(stderr, "Failed to open file to write: %s\nError: %s\n", output_xml_file.c_str(), e.what());
		return -1;
	}

	return 0;
}
