/*
 * appstream-data-composer is an application for mixing multiple
 * appdata and metainfo files into single file
 * Copyright (C) 2019 Aleksei Nikiforov <darktemplar@basealt.ru>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include <stdio.h>

#include <string>
#include <fstream>

#include <boost/program_options.hpp>

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/xml_parser.hpp>
#include <boost/property_tree/ini_parser.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>

void trim_children(boost::property_tree::ptree &propertyTree)
{
	for (auto iter = propertyTree.begin(); iter != propertyTree.end(); ++iter)
	{
		// only trim elements if they become empty after trimming
		auto value = iter->second.get_value<std::string>();
		boost::algorithm::trim(value);

		if (value.empty())
		{
			iter->second.put_value(value);
		}

		trim_children(iter->second);
	}
}

int main(int argc, char **argv)
{
	std::string output_xml_file;
	std::vector<std::string> extra_tags;
	std::vector<std::string> input_xml_file_vector;

	std::set<std::string> processed_files;

	boost::program_options::options_description desc("General Options");
	desc.add_options()
		("help,h", "produce help message")
		("extra_tag", boost::program_options::value<std::vector<std::string> >(&extra_tags), "Additional tag to append to root element. Format is 'tag;value'. May be specified multiple times")
	;

	boost::program_options::options_description hidden("Positional Options");
	hidden.add_options()
		("output_xml_file", boost::program_options::value<std::string>(&output_xml_file)->required(), "Name of file to output resulting data")
		("input_xml_file", boost::program_options::value< std::vector<std::string> >(&input_xml_file_vector)->required(), "Name of file to include into resulting data. May be specified multiple times")
	;

	boost::program_options::options_description all_options;
	all_options.add(desc);
	all_options.add(hidden);

	boost::program_options::positional_options_description pos_opts;

	pos_opts.add("output_xml_file", 1);
	pos_opts.add("input_xml_file", -1);

	boost::program_options::variables_map vm;

	try
	{
		boost::program_options::store(boost::program_options::command_line_parser(argc, argv).options(all_options).positional(pos_opts).run(), vm);
	}
	catch (const std::exception &e)
	{
		fprintf(stderr, "Error: %s\n", e.what());
		return -1;
	}

	if (vm.count("help"))
	{
		std::stringstream desc_stream;
		desc_stream << all_options;

		printf("%s\n", desc_stream.str().c_str());
		return 1;
	}

	try
	{
		boost::program_options::notify(vm);
	}
	catch (const std::exception &e)
	{
		fprintf(stderr, "Error: %s\n", e.what());
		return -1;
	}

	boost::property_tree::ptree outputTree;

	for (auto tag_iter = extra_tags.begin(); tag_iter != extra_tags.end(); ++tag_iter)
	{
		std::vector<std::string> split_vector;
		boost::algorithm::split(split_vector, *tag_iter, boost::is_any_of(";"));

		if (split_vector.size() != 2)
		{
			fprintf(stderr, "Invalid argument: tag = %s\n", tag_iter->c_str());
			return -1;
		}

		outputTree.add_child("<xmlattr>." + split_vector[0], boost::property_tree::ptree(split_vector[1]));
	}

	for (auto iter = input_xml_file_vector.begin(); iter != input_xml_file_vector.end(); ++iter)
	{
		if (processed_files.find(*iter) != processed_files.end())
		{
			continue;
		}

		processed_files.insert(*iter);

		boost::property_tree::ptree inputTree;

		try
		{
			std::ifstream stream(*iter);
			boost::property_tree::read_xml(stream, inputTree);
		}
		catch (const std::exception &e)
		{
			fprintf(stderr, "Failed to open file to read: %s\nError: %s\n", iter->c_str(), e.what());
			return -1;
		}

		trim_children(inputTree);

		auto component_tree = inputTree.get_child_optional("component");
		if (!component_tree)
		{
			fprintf(stderr, "Error: file doesn't have proper root element: %s\n", iter->c_str());
			return -1;
		}

		outputTree.add_child("component", *component_tree);
	}

	try
	{
		boost::property_tree::ptree result;
		result.add_child("components", outputTree);

		std::ofstream out_stream(output_xml_file);
		boost::property_tree::write_xml(out_stream, result, boost::property_tree::xml_writer_make_settings<std::string>('\t', 1));
	}
	catch (const std::exception &e)
	{
		fprintf(stderr, "Failed to open file to write: %s\nError: %s\n", output_xml_file.c_str(), e.what());
		return -1;
	}

	return 0;
}
