/*
 * appstream-data-desktop-converter is an application for creating appdata files from desktop files
 * Copyright (C) 2019 Aleksei Nikiforov <darktemplar@basealt.ru>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include <stdio.h>

#include <string>
#include <fstream>
#include <sstream>
#include <list>
#include <map>
#include <regex>

#include <boost/program_options.hpp>

#include <boost/filesystem/path.hpp>

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/xml_parser.hpp>
#include <boost/property_tree/ini_parser.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>

int main(int argc, char **argv)
{
	std::string input_desktop_file;
	std::string output_xml_file;
	std::string pkgname;
	std::string url;
	std::string metadata_license;
	std::string project_license;
	std::vector<std::string> description;
	std::vector<std::string> releases_vector;
	std::vector<std::string> icons_vector;
	std::vector<std::string> exclude_lang_vector;

	boost::program_options::options_description desc("General Options");
	desc.add_options()
		("help,h", "produce help message")
		("release,r", boost::program_options::value< std::vector<std::string> >(&releases_vector), "add release information to output. format is 'timestamp;version'")
		("icon,i", boost::program_options::value< std::vector<std::string> >(&icons_vector), "add icon information to output. format is 'type;width;height;filename'")
		("exclude_language,e", boost::program_options::value< std::vector<std::string> >(&exclude_lang_vector), "don't include information in specific language")
		("url,u", boost::program_options::value<std::string>(&url), "package homepage url")
		("description,d", boost::program_options::value< std::vector<std::string> >(&description), "description to add to package, if any")
		("metadata_license,m", boost::program_options::value<std::string>(&metadata_license), "generated metadata license")
		("project_license,l", boost::program_options::value<std::string>(&project_license), "project license")
	;

	boost::program_options::options_description hidden("Positional Options");
	hidden.add_options()
		("input_desktop_file", boost::program_options::value<std::string>(&input_desktop_file)->required(), "Input desktop file")
		("output_xml_file", boost::program_options::value<std::string>(&output_xml_file)->required(), "Output appdata.xml file")
		("pkgname", boost::program_options::value<std::string>(&pkgname)->required(), "Name of package containing this appdata.xml file")
	;

	boost::program_options::options_description all_options;
	all_options.add(desc);
	all_options.add(hidden);

	boost::program_options::positional_options_description pos_opts;

	pos_opts.add("input_desktop_file", 1);
	pos_opts.add("output_xml_file", 1);
	pos_opts.add("pkgname", 1);

	boost::program_options::variables_map vm;

	try
	{
		boost::program_options::store(boost::program_options::command_line_parser(argc, argv).options(all_options).positional(pos_opts).run(), vm);
	}
	catch (const std::exception &e)
	{
		fprintf(stderr, "Error: %s\n", e.what());
		return -1;
	}

	if (vm.count("help"))
	{
		std::stringstream desc_stream;
		desc_stream << all_options;

		printf("%s\n", desc_stream.str().c_str());
		return 1;
	}

	try
	{
		boost::program_options::notify(vm);
	}
	catch (const std::exception &e)
	{
		fprintf(stderr, "Error: %s\n", e.what());
		return -1;
	}

	boost::property_tree::ptree propertyTree;

	try
	{
		std::ifstream stream(input_desktop_file);
		boost::property_tree::read_ini(stream, propertyTree);
	}
	catch (const std::exception &e)
	{
		fprintf(stderr, "Failed to open file to read: %s\nError: %s\n", input_desktop_file.c_str(), e.what());
		return -1;
	}

	auto desktop_entry = propertyTree.get_child_optional("Desktop Entry");
	if (!desktop_entry)
	{
		fprintf(stderr, "Error: file %s doesn't contain [Desktop Entry] section\n", input_desktop_file.c_str());
		return -1;
	}

	{
		auto type_element = desktop_entry->get_child_optional("Type");
		if ((!type_element) || (type_element->get_value<std::string>() != "Application"))
		{
			fprintf(stderr, "Error: file %s isn't an 'Application' desktop entry type\n", input_desktop_file.c_str());
			return -1;
		}
	}

	boost::property_tree::ptree applicationPropertyTree;

	applicationPropertyTree.add_child("<xmlattr>.type", boost::property_tree::ptree("desktop"));

	applicationPropertyTree.add_child("id", boost::property_tree::ptree(boost::filesystem::path(input_desktop_file).filename().native()));

	if (!metadata_license.empty())
	{
		applicationPropertyTree.add_child("metadata_license", boost::property_tree::ptree(metadata_license));
	}

	if (!project_license.empty())
	{
		applicationPropertyTree.add_child("project_license", boost::property_tree::ptree(project_license));
	}

	// Process name
	{
		auto ini_element = desktop_entry->get_child_optional("Name");
		if (ini_element)
		{
			applicationPropertyTree.add_child("name", *ini_element);

			// Also check translated items
			std::regex translated_regex("^Name\\[([^\\]]+)\\]$");
			std::smatch results;

			for (auto iter = desktop_entry->begin(); iter != desktop_entry->end(); ++iter)
			{
				if (std::regex_match(iter->first, results, translated_regex))
				{
					if (std::find(exclude_lang_vector.begin(), exclude_lang_vector.end(), results[1].str()) != exclude_lang_vector.end())
					{
						continue;
					}

					boost::property_tree::ptree child(iter->second.get_value<std::string>());
					child.put_child("<xmlattr>.xml:lang", boost::property_tree::ptree(results[1].str()));
					applicationPropertyTree.add_child("name", child);
				}
			}
		}
	}

	// Process comments
	{
		auto ini_element = desktop_entry->get_child_optional("Comment");
		if (ini_element)
		{
			applicationPropertyTree.put_child("summary", *ini_element);

			// Also check translated items
			std::regex translated_regex("^Comment\\[([^\\]]+)\\]$");
			std::smatch results;

			for (auto iter = desktop_entry->begin(); iter != desktop_entry->end(); ++iter)
			{
				if (std::regex_match(iter->first, results, translated_regex))
				{
					if (std::find(exclude_lang_vector.begin(), exclude_lang_vector.end(), results[1].str()) != exclude_lang_vector.end())
					{
						continue;
					}

					boost::property_tree::ptree child(iter->second.get_value<std::string>());
					child.put_child("<xmlattr>.xml:lang", boost::property_tree::ptree(results[1].str()));
					applicationPropertyTree.add_child("summary", child);
				}
			}
		}
	}

	// Process URL if any was passed
	if (!url.empty())
	{
		boost::property_tree::ptree child(url);
		child.put_child("<xmlattr>.type", boost::property_tree::ptree("homepage"));
		applicationPropertyTree.put_child("url", child);
	}

	// Process description if any was passed
	if (!description.empty())
	{
		boost::property_tree::ptree child;

		for (auto iter = description.begin(); iter != description.end(); ++iter)
		{
			child.add_child("p", boost::property_tree::ptree(*iter));
		}

		applicationPropertyTree.put_child("description", child);
	}

	// Process keywords
	{
		auto ini_element = desktop_entry->get_child_optional("Keywords");
		if (ini_element)
		{
			std::vector<std::string> split_vector;
			auto tmpvalue = ini_element->get_value<std::string>();
			boost::algorithm::split(split_vector, tmpvalue, boost::is_any_of(";"));

			bool found = false;
			boost::property_tree::ptree subitem_tree;

			for (auto iter = split_vector.begin(); iter != split_vector.end(); ++iter)
			{
				if (!iter->empty())
				{
					found = true;
					subitem_tree.add_child("keyword", boost::property_tree::ptree(*iter));
				}
			}

			if (found)
			{
				applicationPropertyTree.put_child("keywords", subitem_tree);
			}
		}
	}

	// Process categories
	{
		auto ini_element = desktop_entry->get_child_optional("Categories");
		if (ini_element)
		{
			std::vector<std::string> split_vector;
			auto tmpvalue = ini_element->get_value<std::string>();
			boost::algorithm::split(split_vector, tmpvalue, boost::is_any_of(";"));

			bool found = false;
			boost::property_tree::ptree subitem_tree;

			for (auto iter = split_vector.begin(); iter != split_vector.end(); ++iter)
			{
				if (!iter->empty())
				{
					found = true;
					subitem_tree.add_child("category", boost::property_tree::ptree(*iter));
				}
			}

			if (found)
			{
				applicationPropertyTree.put_child("categories", subitem_tree);
			}
		}
	}

	// Process pkgname
	applicationPropertyTree.put_child("pkgname", boost::property_tree::ptree(pkgname));

	// Process releases
	if (not releases_vector.empty())
	{
		boost::property_tree::ptree releases_tree;

		for (auto iter = releases_vector.begin(); iter != releases_vector.end(); ++iter)
		{
			boost::property_tree::ptree release_item_tree;

			std::vector<std::string> split_vector;
			boost::algorithm::split(split_vector, *iter, boost::is_any_of(";"));

			if (split_vector.size() != 2)
			{
				fprintf(stderr, "Invalid argument: release = %s\n", iter->c_str());
				continue;
			}

			release_item_tree.put_child("<xmlattr>.timestamp", boost::property_tree::ptree(split_vector[0]));
			release_item_tree.put_child("<xmlattr>.version", boost::property_tree::ptree(split_vector[1]));

			releases_tree.add_child("release", release_item_tree);
		}

		applicationPropertyTree.put_child("releases", releases_tree);
	}

	// Process icons
	for (auto iter = icons_vector.begin(); iter != icons_vector.end(); ++iter)
	{
		boost::property_tree::ptree icon_item_tree;

		std::vector<std::string> split_vector;
		boost::algorithm::split(split_vector, *iter, boost::is_any_of(";"));

		if (split_vector.size() != 4)
		{
			fprintf(stderr, "Invalid argument: icon = %s\n", iter->c_str());
			continue;
		}

		icon_item_tree.put_child("<xmlattr>.type", boost::property_tree::ptree(split_vector[0]));
		icon_item_tree.put_child("<xmlattr>.width", boost::property_tree::ptree(split_vector[1]));
		icon_item_tree.put_child("<xmlattr>.height", boost::property_tree::ptree(split_vector[2]));

		icon_item_tree.put_value(split_vector[3]);

		applicationPropertyTree.add_child("icon", icon_item_tree);
	}

	boost::property_tree::ptree outputPropertyTree;

	outputPropertyTree.add_child("component", applicationPropertyTree);

	try
	{
		std::ofstream out_stream(output_xml_file);
		boost::property_tree::write_xml(out_stream, outputPropertyTree, boost::property_tree::xml_writer_make_settings<std::string>('\t', 1));
	}
	catch (const std::exception &e)
	{
		fprintf(stderr, "Failed to open file to write: %s\nError: %s\n", output_xml_file.c_str(), e.what());
		return -1;
	}

	return 0;
}
