/*
 * appstream-data-metainfo-converter is an application for modifying metainfo files
 * Copyright (C) 2019 Aleksei Nikiforov <darktemplar@basealt.ru>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include <stdio.h>

#include <string>
#include <fstream>

#include <boost/program_options.hpp>

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/xml_parser.hpp>
#include <boost/algorithm/string/trim.hpp>

void trim_children(boost::property_tree::ptree &propertyTree)
{
	for (auto iter = propertyTree.begin(); iter != propertyTree.end(); ++iter)
	{
		// only trim elements if they become empty after trimming
		auto value = iter->second.get_value<std::string>();
		boost::algorithm::trim(value);

		if (value.empty())
		{
			iter->second.put_value(value);
		}

		trim_children(iter->second);
	}
}

int main(int argc, char **argv)
{
	std::string input_xml_file;
	std::string output_xml_file;
	std::string pkgname;

	boost::program_options::options_description desc("General Options");
	desc.add_options()
		("help,h", "produce help message")
	;

	boost::program_options::options_description hidden("Positional Options");
	hidden.add_options()
		("input_xml_file", boost::program_options::value<std::string>(&input_xml_file)->required(), "Input metainfo.xml file")
		("output_xml_file", boost::program_options::value<std::string>(&output_xml_file)->required(), "Output metainfo.xml file")
		("pkgname", boost::program_options::value<std::string>(&pkgname)->required(), "Name of package containing this metainfo.xml file")
	;

	boost::program_options::options_description all_options;
	all_options.add(desc);
	all_options.add(hidden);

	boost::program_options::positional_options_description pos_opts;

	pos_opts.add("input_xml_file", 1);
	pos_opts.add("output_xml_file", 1);
	pos_opts.add("pkgname", 1);

	boost::program_options::variables_map vm;

	try
	{
		boost::program_options::store(boost::program_options::command_line_parser(argc, argv).options(all_options).positional(pos_opts).run(), vm);
	}
	catch (const std::exception &e)
	{
		fprintf(stderr, "Error: %s\n", e.what());
		return -1;
	}

	if (vm.count("help"))
	{
		std::stringstream desc_stream;
		desc_stream << all_options;

		printf("%s\n", desc_stream.str().c_str());
		return 1;
	}

	try
	{
		boost::program_options::notify(vm);
	}
	catch (const std::exception &e)
	{
		fprintf(stderr, "Error: %s\n", e.what());
		return -1;
	}

	boost::property_tree::ptree propertyTree;

	try
	{
		std::ifstream stream(input_xml_file);
		boost::property_tree::read_xml(stream, propertyTree);
	}
	catch (const std::exception &e)
	{
		fprintf(stderr, "Failed to open file to read: %s\nError: %s\n", input_xml_file.c_str(), e.what());
		return -1;
	}

	trim_children(propertyTree);

	auto component_tree = propertyTree.get_child_optional("component");
	if (!component_tree)
	{
		fprintf(stderr, "Error: file doesn't have proper root element: %s\n", argv[1]);
		return -1;
	}

	// Process pkgname
	{
		component_tree->erase("pkgname");

		component_tree->put_child("pkgname", boost::property_tree::ptree(pkgname));
	}

	try
	{
		std::ofstream out_stream(output_xml_file);
		boost::property_tree::write_xml(out_stream, propertyTree, boost::property_tree::xml_writer_make_settings<std::string>('\t', 1));
	}
	catch (const std::exception &e)
	{
		fprintf(stderr, "Failed to open file to write: %s\nError: %s\n", output_xml_file.c_str(), e.what());
		return -1;
	}

	return 0;
}
